      SUBROUTINE COVCLC(N, X, F, CV )
*
*     To compute modal covariance matrix for log posterior at mode X.
*
      EXTERNAL F
      INTEGER I, J, IJ, N
      DOUBLE PRECISION X(*), CV(*), F, H, HI, HJ
      PARAMETER ( H = 1D-4 )
      DOUBLE PRECISION FUN, XI, XJ, FPP, FMP, FIP, FMM, FPM, FIM
      FUN = F(X)
      IJ = 0
      DO I = 1,N
         XI = X(I)
         HI = MAX( ABS(XI)*H, H )
         DO J = 1,I-1
            IJ = IJ + 1
            XJ = X(J)
            HJ = MAX( ABS(XJ)*H, H )
            X(I) = XI - HI
            X(J) = XJ - HJ
            FMM = F(X)
            X(J) = XJ + HJ
            FMP = F(X)
            X(I) = XI + HI
            FPP = F(X)
            X(J) = XJ - HJ
            FPM = F(X)
            CV(IJ) = -( FPP-FMP-FPM+FMM )/( 4*HI*HJ )
            X(I) = XI - HI/2
            X(J) = XJ - HJ/2
            FMM = F(X)
            X(J) = XJ + HJ/2
            FMP = F(X)
            X(I) = XI + HI/2
            FPP = F(X)
            X(J) = XJ - HJ/2
            FPM = F(X)
            CV(IJ) = ( -4*( FPP-FMP-FPM+FMM )/( HI*HJ ) - CV(IJ) )/3
            X(J) = XJ
         END DO
         IJ = IJ + 1
         X(I) = XI - HI
         FIM = F(X)
         X(I) = XI + HI
         FIP = F(X)
         CV(IJ) = -( FIP + FIM - 2*FUN )/HI**2
         X(I) = XI - HI/2
         FIM = F(X)
         X(I) = XI + HI/2
         FIP = F(X)
         CV(IJ) = ( -16*( FIP + FIM - 2*FUN )/HI**2 - CV(IJ) )/3
         X(I) = XI
      END DO
      CALL SYMINV( N, CV )
      END
*
      SUBROUTINE SYMINV( N, LOWINV )
*
*     Computes lower symmetric inverse and determinant in situ
*
      INTEGER N
      DOUBLE PRECISION LOWINV(*)
      CALL CHOLSK(N, LOWINV)
      CALL CHOLNV(N, LOWINV)
      CALL CHOLPI(N, LOWINV)
      END
      SUBROUTINE CHOLSK( N, CHOFAC )
*
*     Computes Choleski factor in situ
*
      INTEGER I, II, J, JJ, K, N
      DOUBLE PRECISION CHOFAC(*), T, S, ZERO
      PARAMETER ( ZERO = 0 )
      JJ = 0
      DO J = 1,N
         II = JJ
         DO I = J,N
            S = CHOFAC(II+J)
            DO K = 1,J-1
               S = S - CHOFAC(II+K)*CHOFAC(JJ+K)
            END DO
            IF ( I .EQ. J ) THEN
               T = SQRT( MAX( S, ZERO ) )
               CHOFAC(II+J) = T
            ELSE IF ( T .GT. 0 ) THEN
               CHOFAC(II+J) = S/T
            ELSE
               CHOFAC(II+J) = 0
            ENDIF
            II = II + I
         END DO
         JJ = JJ + J
      END DO
      END
      SUBROUTINE CHOLNV(N, CHOINV)
*
*     Inverts a lower triangular matrix in situ
*
      INTEGER I, II, J, JJ, K, KK, N
      DOUBLE PRECISION CHOINV(*), T
      DOUBLE PRECISION S
      II = 0
      DO I = 1,N
         T = 1/CHOINV(II+I)
         JJ = 0
         DO J = 1,I-1
            S = 0
            JJ = JJ + J
            KK = JJ
            DO K = J,I-1
               S = S + CHOINV(II+K)*CHOINV(KK)
               KK = KK + K
            END DO
            CHOINV(II+J) = -S*T
         END DO
         II = II + I
         CHOINV(II) = T
      END DO
      END
      SUBROUTINE CHOLPD(N, CHOPRD)
*
*     Multiplies Choleski factors in situ
*
      INTEGER I, II, J, K, KK, N, NN
      DOUBLE PRECISION CHOPRD(*)
      DOUBLE PRECISION S
      NN = (N*(N+1))/2
      KK = NN
      DO K = N,1,-1
         KK = KK - K
         II = NN
         DO I = N,K,-1
            II = II - I
            S = 0
            DO J = 1,K
               S = S + CHOPRD(II+J)*CHOPRD(KK+J)
            END DO
            CHOPRD(II+K) = S
         END DO
      END DO
      END
      SUBROUTINE CHOLPI(N, CHOPDI)
*
*     Multiplies Choleski inverse factors in situ
*
      INTEGER I, II, J, JJ, K, KK, N
      DOUBLE PRECISION CHOPDI(*)
      DOUBLE PRECISION S
      II = 0
      DO I = 1,N
         DO J = 1,I
            S = 0
            JJ = II + I
            KK = II + J
            DO K = I,N
               S = S + CHOPDI(KK)*CHOPDI(JJ)
               JJ = JJ + K
               KK = KK + K
            END DO
            CHOPDI(II+J) = S
         END DO
         II = II + I
      END DO
      END
